@@ -1,5 +1,5 @@
 /* stat.c -- display file or file system status
-   Copyright (C) 2001-2021 Free Software Foundation, Inc.
+   Copyright (C) 2001-2022 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -114,7 +114,8 @@ static bool append_exe;
 /* BeOS has a statvfs function, but it does not return sensible values
    for f_files, f_ffree and f_favail, and lacks f_type, f_basetype and
    f_fstypename.  Use 'struct fs_info' instead.  */
-static int ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static int
 statfs (char const *filename, struct fs_info *buf)
 {
   dev_t device = dev_for_path (filename);
@@ -266,7 +267,8 @@ print_stat (char *pformat, size_t prefix_len, char mod, char m,
    Others have statfs.f_fstypename[MFSNAMELEN] (NetBSD 1.5.2).
    Still others have neither and have to get by with f_type (GNU/Linux).
    But f_type may only exist in statfs (Cygwin).  */
-static char const * ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static char const *
 human_fstype (STRUCT_STATVFS const *statfsbuf)
 {
 #ifdef STATXFS_FILE_SYSTEM_TYPE_MEMBER_NAME
@@ -641,7 +643,8 @@ human_fstype (STRUCT_STATVFS const *statfsbuf)
 #endif
 }
 
-static char * ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static char *
 human_access (struct stat const *statbuf)
 {
   static char modebuf[12];
@@ -650,7 +653,8 @@ human_access (struct stat const *statbuf)
   return modebuf;
 }
 
-static char * ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static char *
 human_time (struct timespec t)
 {
   /* STR must be at least INT_BUFSIZE_BOUND (intmax_t) big, either
@@ -840,7 +844,8 @@ out_epoch_sec (char *pformat, size_t prefix_len,
 
 /* Print the context information of FILENAME, and return true iff the
    context could not be obtained.  */
-static bool ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static bool
 out_file_context (char *pformat, size_t prefix_len, char const *filename)
 {
   char *scontext;
@@ -863,8 +868,9 @@ out_file_context (char *pformat, size_t prefix_len, char const *filename)
 }
 
 /* Print statfs info.  Return zero upon success, nonzero upon failure.  */
-static bool ATTRIBUTE_WARN_UNUSED_RESULT
-print_statfs (char *pformat, size_t prefix_len, char mod _GL_UNUSED, char m,
+NODISCARD
+static bool
+print_statfs (char *pformat, size_t prefix_len, MAYBE_UNUSED char mod, char m,
               int fd, char const *filename,
               void const *data)
 {
@@ -951,7 +957,8 @@ print_statfs (char *pformat, size_t prefix_len, char mod _GL_UNUSED, char m,
 /* Return any bind mounted source for a path.
    The caller should not free the returned buffer.
    Return NULL if no bind mount found.  */
-static char const * ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static char const *
 find_bind_mount (char const * name)
 {
   char const * bind_mount = NULL;
@@ -990,7 +997,8 @@ find_bind_mount (char const * name)
 }
 
 /* Print mount point.  Return zero upon success, nonzero upon failure.  */
-static bool ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static bool
 out_mount_point (char const *filename, char *pformat, size_t prefix_len,
                  const struct stat *statp)
 {
@@ -1115,8 +1123,9 @@ print_esc_char (char c)
   putchar (c);
 }
 
-static size_t _GL_ATTRIBUTE_PURE
-format_code_offset (char const* directive)
+ATTRIBUTE_PURE
+static size_t
+format_code_offset (char const *directive)
 {
   size_t len = strspn (directive + 1, printf_flags);
   char const *fmt_char = directive + len + 1;
@@ -1129,7 +1138,8 @@ format_code_offset (char const* directive)
 /* Print the information specified by the format string, FORMAT,
    calling PRINT_FUNC for each %-directive encountered.
    Return zero upon success, nonzero upon failure.  */
-static bool ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static bool
 print_it (char const *format, int fd, char const *filename,
           bool (*print_func) (char *, size_t, char, char,
                               int, char const *, void const *),
@@ -1257,7 +1267,8 @@ print_it (char const *format, int fd, char const *filename,
 }
 
 /* Stat the file system and print what we find.  */
-static bool ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static bool
 do_statfs (char const *filename, char const *format)
 {
   STRUCT_STATVFS statfsbuf;
@@ -1356,7 +1367,8 @@ fmt_to_mask (char fmt)
   return 0;
 }
 
-static unsigned int _GL_ATTRIBUTE_PURE
+ATTRIBUTE_PURE
+static unsigned int
 format_to_mask (char const *format)
 {
   unsigned int mask = 0;
@@ -1376,7 +1388,8 @@ format_to_mask (char const *format)
 }
 
 /* statx the file and print what we find */
-static bool ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static bool
 do_stat (char const *filename, char const *format, char const *format2)
 {
   int fd = STREQ (filename, "-") ? 0 : AT_FDCWD;
@@ -1403,6 +1416,9 @@ do_stat (char const *filename, char const *format, char const *format2)
   else if (force_sync)
     flags |= AT_STATX_FORCE_SYNC;
 
+  if (! force_sync)
+    flags |= AT_NO_AUTOMOUNT;
+
   fd = statx (fd, pathname, flags, format_to_mask (format), &stx);
   if (fd < 0)
     {
@@ -1460,7 +1476,8 @@ get_birthtime (int fd, char const *filename, struct stat const *st)
 
 
 /* stat the file and print what we find */
-static bool ATTRIBUTE_WARN_UNUSED_RESULT
+NODISCARD
+static bool
 do_stat (char const *filename, char const *format,
          char const *format2)
 {
@@ -1995,5 +2012,5 @@ main (int argc, char *argv[])
            ? do_statfs (argv[i], format)
            : do_stat (argv[i], format, format2));
 
-  return ok ? EXIT_SUCCESS : EXIT_FAILURE;
+  main_exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
